#!/bin/bash

# Check if no arguments are given
if [ $# -eq 0 ]; then
  echo "Usage: ./run_tests.sh [-s] [solver binary with arguments...]"
  echo "       -s --> only short test run (with around 90 instances) is executed"
  echo "              for Pacose less than 2 seconds run time"
  exit 1
fi

# logfolder="/tmp/Exact/$2"
if [ "$1" = "-s" ]; then
  long=false
  binary="${*:2}"
else
  long=true
  binary="${*:1}"
fi
# options="--timeout=$1 --test=1 $4"


SCRIPTPATH="$( cd "$(dirname "$0")" >/dev/null 2>&1 ; pwd -P )"

errors=0
modelErrors=0
emptyOvalue=0
ovalueErrors=0
tested=0
completed=0

echo "###########################"
echo "########## START ##########"
echo "###########################"
echo ""
#echo "data: $logfolder"
echo "binary: $binary"
#echo "options: $options"
echo ""

if [ "$1" = "-s" ]; then
declare -a wcnfArray=(
"baseWCNFs/emptyClause.wcnf*UNSATISFIABLE"
"baseWCNFs/emptySoftClause.wcnf*1"
"baseWCNFs/emptySoftClauses.wcnf*3"
"baseWCNFs/emptySoftClausesWithHardClauses.wcnf*3"
"baseWCNFs/emptySoftClauseWithUnsatHardClauses.wcnf*UNSATISFIABLE"
"baseWCNFs/emptySoftClauseWithNormalSoftClauseWithHardClauses.wcnf*6"
"baseWCNFs/empty.wcnf*0"
"baseWCNFs/MinimalUnsat.wcnf*UNSATISFIABLE"
"baseWCNFs/OneHardUnitDoesNotContainLiteralOne.wcnf*0"
"baseWCNFs/OneHardUnit.wcnf*0"
"baseWCNFs/OneSoftUnitWeight1.wcnf*0"
"baseWCNFs/OneSoftUnitWeightUINT32Maxplus1.wcnf*0"
"baseWCNFs/smallo0.wcnf*0"
"baseWCNFs/smallo1.wcnf*1"
"baseWCNFs/TautologyHardClause.wcnf*0"
"baseWCNFs/TautologySoftClause.wcnf*0"
"baseWCNFs/TwoMinimalContradictingSoftClauses.wcnf*1"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_12_red-1365857429670963003.wcnf*0"
#"reducedUniqueWCNFs/Cashwmaxsat-Plus_14_red-4543078358654527564.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_18_red-2839957876382344943.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_42_red-2100714977247962924.wcnf*1"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_44_red-2258465440550850344.wcnf*13215994667762143"
#"reducedUniqueWCNFs/Cashwmaxsat-Plus_45_red-4143592419272789868.wcnf*8191932373021177"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_48_red-3584631550324119099.wcnf*924295644"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_50_red-4359572950483299044.wcnf*0"
#"reducedUniqueWCNFs/Cashwmaxsat-Plus_56_red-1048208781981639312.wcnf*4029182931969790"
#"reducedUniqueWCNFs/Cashwmaxsat-Plus_59_red-1270251658374345229.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_9_red-812881557716518233.wcnf*1"
#"reducedUniqueWCNFs/CGSS_18_red-2292127357992646614.wcnf*7727"
"reducedUniqueWCNFs/CGSS_42_red-1487924879221147942.wcnf*40707926424544505"
#"reducedUniqueWCNFs/CGSS_44_red-1224009853503763696.wcnf*14760652175744823"
"reducedUniqueWCNFs/CGSS_45_red-1012855063300641881.wcnf*12167739843464173"
#"reducedUniqueWCNFs/CGSS_48_red-1206691116550062767.wcnf*383517261"
#"reducedUniqueWCNFs/Exact_18_red-4025587915001303986.wcnf*317442"
#"reducedUniqueWCNFs/Exact_1_red-1611952973005558687.wcnf*190300949"
"reducedUniqueWCNFs/Exact_31_red-116854817381655133.wcnf*294920058"
"reducedUniqueWCNFs/MaxHS_12_red-182969811668377320.wcnf*0"
#"reducedUniqueWCNFs/MaxHS_13_red-4521713885359821480.wcnf*0"
"reducedUniqueWCNFs/MaxHS_14_red-1142333688832469272.wcnf*0"
"reducedUniqueWCNFs/MaxHS_14_red-2378267317391590079.wcnf*0"
#"reducedUniqueWCNFs/MaxHS_15_red-4471997378323133042.wcnf*1"
"reducedUniqueWCNFs/MaxHS_18_red-1437300044257259909.wcnf*1"
"reducedUniqueWCNFs/MaxHS_1_red-3199522118874382216.wcnf*2"
#"reducedUniqueWCNFs/MaxHS_23_red-1369228926978153807.wcnf*127"
"reducedUniqueWCNFs/MaxHS_2_red-3708664613699879098.wcnf*12"
#"reducedUniqueWCNFs/MaxHS_30_red-1636217099346116043.wcnf*45"
#"reducedUniqueWCNFs/MaxHS_31_red-3537275520299263569.wcnf*4429674011"
"reducedUniqueWCNFs/MaxHS_32_red-496328484390903258.wcnf*942862119"
"reducedUniqueWCNFs/MaxHS_42_red-576324400294307223.wcnf*844836753"
"reducedUniqueWCNFs/MaxHS_43_red-113106383842623885.wcnf*1719880153"
#"reducedUniqueWCNFs/MaxHS_44_red-2931653447379948941.wcnf*21849752967009851"
"reducedUniqueWCNFs/MaxHS_45_red-1634190298382386325.wcnf*114674111"
"reducedUniqueWCNFs/MaxHS_48_red-3618368612976155930.wcnf*923990351"
#"reducedUniqueWCNFs/MaxHS_53_red-3273341903972678006.wcnf*1300571833"
#"reducedUniqueWCNFs/MaxHS_60_red-3683434563126480531.wcnf*306885689"
"reducedUniqueWCNFs/maxpre2EvalMaxSAT_13_red-1787893125344086754.wcnf*1"
"reducedUniqueWCNFs/maxpre2EvalMaxSAT_39_red-1964975267777150992.wcnf*1"
#"reducedUniqueWCNFs/maxpre2EvalMaxSAT_43_red-2868335317395646542.wcnf*2"
"reducedUniqueWCNFs/maxpre2EvalMaxSAT_56_red-4599407379028638546.wcnf*0"
"reducedUniqueWCNFs/maxpre2EvalMaxSAT_9_red-340722805977130735.wcnf*0"
#"reducedUniqueWCNFs/Pacose_12_red-408069156423875169.wcnf*0"
"reducedUniqueWCNFs/Pacose_24_red-1286162711923734785.wcnf*17"
"reducedUniqueWCNFs/Pacose_26_red-3988839199503314461.wcnf*0"
"reducedUniqueWCNFs/Pacose_30_red-2167694050811453563.wcnf*36"
"reducedUniqueWCNFs/Pacose_35_red-2951893180766256529.wcnf*91398525"
"reducedUniqueWCNFs/Pacose_54_red-1358157514203883747.wcnf*26426196704"
"reducedUniqueWCNFs/Pacose_56_red-3281457200392413701.wcnf*2223671546126634343"
"reducedUniqueWCNFs/Pacose_5_red-3802828968457918065.wcnf*54047"
"reducedUniqueWCNFs/UWrMaxSat_18_red-434222651361422376.wcnf*62"
#"reducedUniqueWCNFs/UWrMaxSat_48_red-3345226870336081251.wcnf*2666215797507687042"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_12_red-708605192798371900.wcnf*1"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_14_red-1910526412935573211.wcnf*0"
#"reducedUniqueWCNFs/UWrMaxSat-SCIP_18_red-3728880575233737753.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_39_red-185185273772798505.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_42_red-2206054070726780272.wcnf*945747779"
#"reducedUniqueWCNFs/UWrMaxSat-SCIP_44_red-3281457200392413701.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_45_red-4342119793509242537.wcnf*1"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_48_red-1093108851438549088.wcnf*1650022201"
#"reducedUniqueWCNFs/UWrMaxSat-SCIP_50_red-719310437759088578.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_59_red-1270251658374345229.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_9_red-827404868206741790.wcnf*UNSATISFIABLE"
#"reducedUniqueWCNFs/WMaxCDCL_31_red-2853693958521466174.wcnf*16906386530742218"
"reducedUniqueWCNFs/WMaxCDCL_39_red-3760685081229858521.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/WMaxCDCL_40_red-1230756892310092431.wcnf*3380361619521023"
#"reducedUniqueWCNFs/WMaxCDCL_42_red-353414103522320368.wcnf*9491922604943463"
"reducedUniqueWCNFs/WMaxCDCL_44_red-3247157704190641489.wcnf*14113912232380779"
"reducedUniqueWCNFs/WMaxCDCL_45_red-4504257650829621829.wcnf*9319692811377733"
#"reducedUniqueWCNFs/WMaxCDCL_48_red-2726655741059448275.wcnf*22201918151458737"
"reducedUniqueWCNFs/WMaxCDCL_60_red-3269417686496778418.wcnf*356393726267587704"
"reducedUniqueWCNFs/WMaxCDCL_9_red-960852779331877826.wcnf*UNSATISFIABLE"
#"reducedUniqueWCNFs/WMaxCDCL-BandAll_14_red-1573595701574387285.wcnf*2"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_15_red-4464408172294026275.wcnf*5"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_31_red-1650492680461762036.wcnf*9094379793822221"
#"reducedUniqueWCNFs/WMaxCDCL-BandAll_34_red-3910947893840279638.wcnf*8145900349697497"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_39_red-1130975699679110928.wcnf*UNSATISFIABLE"
#"reducedUniqueWCNFs/WMaxCDCL-BandAll_40_red-4413781331871572532.wcnf*1329398780228330"
#"reducedUniqueWCNFs/WMaxCDCL-BandAll_42_red-1016238135184499311.wcnf*2"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_44_red-2672274648256964035.wcnf*9125479794677607"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_45_red-2258465440550850344.wcnf*12389995001027008"
#"reducedUniqueWCNFs/WMaxCDCL-BandAll_48_red-832511782895553194.wcnf*242000448646113205"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_56_red-4261954262327454570.wcnf*237803482746857698"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_60_red-786798723718669031.wcnf*4157178231546598"
#"reducedUniqueWCNFs/WMaxCDCL-BandAll_9_red-2042077545131207184.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/z3maxres_12_red-4442551375815625515.wcnf*4"
"reducedUniqueWCNFs/z3maxres_18_red-1315352413729120633.wcnf*1"
#"reducedUniqueWCNFs/z3maxres_31_red-30358273458052886.wcnf*1882211714"
"reducedUniqueWCNFs/z3maxres_40_red-1682766907044057079.wcnf*2906095073836"
"reducedUniqueWCNFs/z3maxres_42_red-786798723718669031.wcnf*508967635"
"reducedUniqueWCNFs/z3maxres_45_red-1544036629320291585.wcnf*4453186682"
#"reducedUniqueWCNFs/z3maxres_48_red-514898981072592001.wcnf*2160415111"
"reducedUniqueWCNFs/z3rc2_12_red-2296016229024490956.wcnf*2"
"reducedUniqueWCNFs/z3rc2_15_red-1559107368525074383.wcnf*3"
"reducedUniqueWCNFs/z3rc2_18_red-175513428354106618.wcnf*2"
#"reducedUniqueWCNFs/z3rc2_40_red-4504257650829621829.wcnf*8594142881685"
"reducedUniqueWCNFs/z3rc2_42_red-2030875441587988817.wcnf*1772420315"
"reducedUniqueWCNFs/z3rc2_45_red-946033746336718185.wcnf*4305529855"
#"reducedUniqueWCNFs/z3rc2_48_red-2939344604895055236.wcnf*465872227"
"reducedUniqueWCNFs/z3wmax_23_red-4520349575283451183.wcnf*0"
"reducedUniqueWCNFs/z3wmax_39_red-1957787610028959388.wcnf*0"
#"reducedUniqueWCNFs/z3wmax_40_red-2735645343719678363.wcnf*1482755547"
"reducedUniqueWCNFs/z3wmax_42_red-1639582119976648487.wcnf*46022862304192"
"reducedUniqueWCNFs/z3wmax_43_red-419516046229911735.wcnf*0"
#"reducedUniqueWCNFs/z3wmax_45_red-2161187820670240308.wcnf*4572868525"
"reducedUniqueWCNFs/z3wmax_48_red-3604028936772614171.wcnf*1243373556"
"reducedUniqueWCNFs/z3wmax_53_red-1084762891037370712.wcnf*0"
)
else
declare -a wcnfArray=(
"mseWCNFs/driverlog01bc.wcsp.dir.wcnf*2245"
"mseWCNFs/normalized-factor-size=9-P=383-Q=509.opb.wcnf*383"
"mseWCNFs/WCNF_pathways_p02.wcnf*3"
"mseWCNFs/archlinux.dimacs.wcnf*11744"
#"mseWCNFs/414.wcsp.dir.wcnf*38478"
#"mseWCNFs/g2_n56e57_n61e85.wcnf*39"
"mseWCNFs/synthetic-20.wcnf*6720"
#"mseWCNFs/f49-DC_TotalLoss.seq-A-2-2-EDCBAir.wcnf*85435860665"
"baseWCNFs/emptyClause.wcnf*UNSATISFIABLE"
"baseWCNFs/emptySoftClause.wcnf*1"
"baseWCNFs/emptySoftClauses.wcnf*3"
"baseWCNFs/emptySoftClausesWithHardClauses.wcnf*3"
"baseWCNFs/emptySoftClauseWithUnsatHardClauses.wcnf*UNSATISFIABLE"
"baseWCNFs/emptySoftClauseWithNormalSoftClauseWithHardClauses.wcnf*6"
"baseWCNFs/empty.wcnf*0"
"baseWCNFs/MinimalUnsat.wcnf*UNSATISFIABLE"
"baseWCNFs/OneHardUnitDoesNotContainLiteralOne.wcnf*0"
"baseWCNFs/OneHardUnit.wcnf*0"
"baseWCNFs/OneSoftUnitWeight1.wcnf*0"
"baseWCNFs/OneSoftUnitWeightUINT32Maxplus1.wcnf*0"
"baseWCNFs/smallo0.wcnf*0"
"baseWCNFs/smallo1.wcnf*1"
"baseWCNFs/TautologyHardClause.wcnf*0"
"baseWCNFs/TautologySoftClause.wcnf*0"
"baseWCNFs/TwoMinimalContradictingSoftClauses.wcnf*1"
"reducedUniqueWCNFs/Cashwmaxsat-CorePlus_12_red-1520613405909291775.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-CorePlus_12_red-2262641803525952700.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-CorePlus_12_red-3938664394182135786.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-CorePlus_14_red-2603027783804086454.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-CorePlus_14_red-4082988606349055981.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-CorePlus_14_red-623470715727282899.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-CorePlus_18_red-1849880863928801000.wcnf*3"
"reducedUniqueWCNFs/Cashwmaxsat-CorePlus_18_red-2448297982864257022.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-CorePlus_18_red-2510077496917081429.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-CorePlus_18_red-596327489556597172.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-CorePlus_42_red-1136848343093114422.wcnf*1"
"reducedUniqueWCNFs/Cashwmaxsat-CorePlus_42_red-2625136892140343246.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-CorePlus_42_red-773616030500625105.wcnf*3138411830997198"
"reducedUniqueWCNFs/Cashwmaxsat-CorePlus_44_red-2173826577747340621.wcnf*12107738582301435"
"reducedUniqueWCNFs/Cashwmaxsat-CorePlus_44_red-4466927197129988359.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-CorePlus_45_red-274178739103012090.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-CorePlus_45_red-4113462657865614335.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-CorePlus_45_red-4413781331871572532.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-CorePlus_45_red-4504257650829621829.wcnf*9701968443222049"
"reducedUniqueWCNFs/Cashwmaxsat-CorePlus_48_red-2097032348476744928.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-CorePlus_48_red-2887089214911469495.wcnf*1169926422"
"reducedUniqueWCNFs/Cashwmaxsat-CorePlus_48_red-3545243853613679243.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-CorePlus_56_red-1048208781981639312.wcnf*4029182931969790"
"reducedUniqueWCNFs/Cashwmaxsat-CorePlus_59_red-1065452161282436221.wcnf*1"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_12_red-1365857429670963003.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_12_red-1664189666936937288.wcnf*1"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_12_red-1955130124151887222.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_12_red-420343497460491473.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_14_red-2366835664028298576.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_14_red-4339406619625695919.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_14_red-4543078358654527564.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_18_red-2839957876382344943.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_18_red-2840871483796178012.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_18_red-293586797620349920.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_18_red-3130281095349926019.wcnf*3"
#"reducedUniqueWCNFs/Cashwmaxsat-Plus_33_red-1069423359933276817.wcnf*9239063301820962052" # sum of weights > 2^63
#"reducedUniqueWCNFs/Cashwmaxsat-Plus_33_red-2320692503609130440.wcnf*9233114895876828182" # sum of weights > 2^63
#"reducedUniqueWCNFs/Cashwmaxsat-Plus_33_red-4489510176210977570.wcnf*9267979088012949242" # sum of weights > 2^63
#"reducedUniqueWCNFs/Cashwmaxsat-Plus_33_red-4604869691953281237.wcnf*9229160939207881892" # sum of weights > 2^63
#"reducedUniqueWCNFs/Cashwmaxsat-Plus_33_red-864786421590678020.wcnf*9269297536667740589" # sum of weights > 2^63
"reducedUniqueWCNFs/Cashwmaxsat-Plus_42_red-1411231300624613679.wcnf*533936702"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_42_red-179286200279657491.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_42_red-1974616838568923064.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_42_red-2100714977247962924.wcnf*1"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_44_red-2258465440550850344.wcnf*13215994667762143"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_44_red-2931653447379948941.wcnf*18144455427694870"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_44_red-3545938599546565470.wcnf*8810669284777143"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_44_red-582486669401176224.wcnf*11466999516137531"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_45_red-2037273550548958332.wcnf*12218079392565673"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_45_red-3281457200392413701.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_45_red-3551257043634425947.wcnf*1"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_45_red-4143592419272789868.wcnf*8191932373021177"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_48_red-3125707873862232604.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_48_red-3584631550324119099.wcnf*924295644"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_48_red-3773895843148146771.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_48_red-4531733537433352914.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_50_red-1249063163625476418.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_50_red-3827889383800860723.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_50_red-4359572950483299044.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_56_red-1048208781981639312.wcnf*4029182931969790"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_59_red-1270251658374345229.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_59_red-1985418537756727306.wcnf*1"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_59_red-2743116627831808271.wcnf*1"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_59_red-3587769345544155790.wcnf*0"
"reducedUniqueWCNFs/Cashwmaxsat-Plus_9_red-812881557716518233.wcnf*1"
"reducedUniqueWCNFs/CGSS_18_red-2292127357992646614.wcnf*7727"
"reducedUniqueWCNFs/CGSS_18_red-2727021659120573577.wcnf*1842"
"reducedUniqueWCNFs/CGSS_18_red-3027352830162997443.wcnf*2920"
"reducedUniqueWCNFs/CGSS_18_red-4155975368981508091.wcnf*468"
"reducedUniqueWCNFs/CGSS_42_red-1487924879221147942.wcnf*40707926424544505"
"reducedUniqueWCNFs/CGSS_42_red-3021608191358524543.wcnf*32374453385912761"
"reducedUniqueWCNFs/CGSS_42_red-745085082311664877.wcnf*38915191643324629"
"reducedUniqueWCNFs/CGSS_44_red-1224009853503763696.wcnf*14760652175744823"
"reducedUniqueWCNFs/CGSS_44_red-196753209117644832.wcnf*20437330456252702"
"reducedUniqueWCNFs/CGSS_44_red-2258465440550850344.wcnf*11186837941963895"
"reducedUniqueWCNFs/CGSS_44_red-4502739501875098556.wcnf*32255259136799167"
"reducedUniqueWCNFs/CGSS_45_red-1012855063300641881.wcnf*12167739843464173"
"reducedUniqueWCNFs/CGSS_45_red-2868335317395646542.wcnf*9920581098195025"
"reducedUniqueWCNFs/CGSS_45_red-3537275520299263569.wcnf*10098283361633821"
"reducedUniqueWCNFs/CGSS_48_red-1206691116550062767.wcnf*383517261"
"reducedUniqueWCNFs/CGSS_48_red-3254119129266484915.wcnf*1646630784"
"reducedUniqueWCNFs/CGSS_48_red-4480925500239101576.wcnf*221313517"
"reducedUniqueWCNFs/CGSS_48_red-5770998853622781.wcnf*427409550"
"reducedUniqueWCNFs/Exact_18_red-4025587915001303986.wcnf*317442"
"reducedUniqueWCNFs/Exact_1_red-1611952973005558687.wcnf*190300949"
"reducedUniqueWCNFs/Exact_1_red-1795142913688699408.wcnf*0"
"reducedUniqueWCNFs/Exact_1_red-2386568577569611393.wcnf*62551716"
"reducedUniqueWCNFs/Exact_31_red-116854817381655133.wcnf*294920058"
"reducedUniqueWCNFs/Exact_31_red-2635857077823969549.wcnf*167382737"
"reducedUniqueWCNFs/MaxHS_12_red-182969811668377320.wcnf*0"
"reducedUniqueWCNFs/MaxHS_12_red-2367393121091382333.wcnf*1"
"reducedUniqueWCNFs/MaxHS_12_red-3583838767742781403.wcnf*6"
"reducedUniqueWCNFs/MaxHS_12_red-4019153774600098984.wcnf*0"
"reducedUniqueWCNFs/MaxHS_13_red-2406752858549471466.wcnf*0"
"reducedUniqueWCNFs/MaxHS_13_red-3956046307739898430.wcnf*0"
"reducedUniqueWCNFs/MaxHS_13_red-4516936184578985642.wcnf*0"
"reducedUniqueWCNFs/MaxHS_13_red-4521713885359821480.wcnf*0"
"reducedUniqueWCNFs/MaxHS_14_red-1142333688832469272.wcnf*0"
"reducedUniqueWCNFs/MaxHS_14_red-1904774257588630617.wcnf*0"
"reducedUniqueWCNFs/MaxHS_14_red-2378267317391590079.wcnf*0"
"reducedUniqueWCNFs/MaxHS_14_red-3733304339593216197.wcnf*0"
"reducedUniqueWCNFs/MaxHS_15_red-1439269513515557125.wcnf*1"
"reducedUniqueWCNFs/MaxHS_15_red-2889652885631822396.wcnf*1"
"reducedUniqueWCNFs/MaxHS_15_red-413349474436735275.wcnf*1"
"reducedUniqueWCNFs/MaxHS_15_red-4471997378323133042.wcnf*1"
"reducedUniqueWCNFs/MaxHS_18_red-1437300044257259909.wcnf*1"
"reducedUniqueWCNFs/MaxHS_18_red-299582112903134320.wcnf*0"
"reducedUniqueWCNFs/MaxHS_18_red-3520127812268797183.wcnf*1"
"reducedUniqueWCNFs/MaxHS_18_red-868802023606758978.wcnf*0"
"reducedUniqueWCNFs/MaxHS_1_red-1674876392373026900.wcnf*1"
"reducedUniqueWCNFs/MaxHS_1_red-193251431004265909.wcnf*1"
"reducedUniqueWCNFs/MaxHS_1_red-3199522118874382216.wcnf*2"
"reducedUniqueWCNFs/MaxHS_23_red-1369228926978153807.wcnf*127"
"reducedUniqueWCNFs/MaxHS_23_red-1536681433119686559.wcnf*37"
"reducedUniqueWCNFs/MaxHS_23_red-256625930015361976.wcnf*43"
"reducedUniqueWCNFs/MaxHS_23_red-2802022998262041612.wcnf*47"
"reducedUniqueWCNFs/MaxHS_2_red-3708664613699879098.wcnf*12"
"reducedUniqueWCNFs/MaxHS_30_red-1636217099346116043.wcnf*45"
"reducedUniqueWCNFs/MaxHS_30_red-3124702702487181450.wcnf*40"
"reducedUniqueWCNFs/MaxHS_30_red-3353865968981105023.wcnf*41"
"reducedUniqueWCNFs/MaxHS_30_red-3382409651942168553.wcnf*45"
"reducedUniqueWCNFs/MaxHS_31_red-1181304565653013676.wcnf*1968006452"
"reducedUniqueWCNFs/MaxHS_31_red-1398515925330311500.wcnf*1639534535"
"reducedUniqueWCNFs/MaxHS_31_red-3537275520299263569.wcnf*4429674011"
"reducedUniqueWCNFs/MaxHS_32_red-496328484390903258.wcnf*942862119"
"reducedUniqueWCNFs/MaxHS_42_red-1043977331583610794.wcnf*859516190"
"reducedUniqueWCNFs/MaxHS_42_red-1246120152132078020.wcnf*185691149"
"reducedUniqueWCNFs/MaxHS_42_red-2203871708166975328.wcnf*1507208518"
"reducedUniqueWCNFs/MaxHS_42_red-576324400294307223.wcnf*844836753"
"reducedUniqueWCNFs/MaxHS_43_red-113106383842623885.wcnf*1719880153"
"reducedUniqueWCNFs/MaxHS_43_red-2888032149468054567.wcnf*1331922448"
"reducedUniqueWCNFs/MaxHS_43_red-2923434174059892951.wcnf*0"
"reducedUniqueWCNFs/MaxHS_43_red-3879298181394504739.wcnf*179803801"
"reducedUniqueWCNFs/MaxHS_44_red-2052882519405773055.wcnf*380958632"
"reducedUniqueWCNFs/MaxHS_44_red-2317167705721969343.wcnf*12923786008107735"
"reducedUniqueWCNFs/MaxHS_44_red-2370797552336548264.wcnf*9091552615103215"
"reducedUniqueWCNFs/MaxHS_44_red-2931653447379948941.wcnf*21849752967009851"
"reducedUniqueWCNFs/MaxHS_45_red-1634190298382386325.wcnf*114674111"
"reducedUniqueWCNFs/MaxHS_45_red-2173826577747340621.wcnf*10124173013682493"
"reducedUniqueWCNFs/MaxHS_45_red-4315342728986121571.wcnf*9977322536673301"
"reducedUniqueWCNFs/MaxHS_45_red-4502770299894597602.wcnf*11670476817261597"
"reducedUniqueWCNFs/MaxHS_48_red-1953277059597663916.wcnf*1243211573"
"reducedUniqueWCNFs/MaxHS_48_red-231831315273589144.wcnf*699175178"
"reducedUniqueWCNFs/MaxHS_48_red-3276841890134449563.wcnf*0"
"reducedUniqueWCNFs/MaxHS_48_red-3618368612976155930.wcnf*923990351"
"reducedUniqueWCNFs/MaxHS_53_red-3273341903972678006.wcnf*1300571833"
"reducedUniqueWCNFs/MaxHS_53_red-3731508303756772707.wcnf*1583053844"
"reducedUniqueWCNFs/MaxHS_53_red-4357514472202144084.wcnf*951739331"
"reducedUniqueWCNFs/MaxHS_53_red-823594533813448437.wcnf*785032992"
"reducedUniqueWCNFs/MaxHS_60_red-2224886831523908799.wcnf*598361046"
"reducedUniqueWCNFs/MaxHS_60_red-2283355287300457314.wcnf*1040129776"
"reducedUniqueWCNFs/MaxHS_60_red-3355069092955409388.wcnf*318550042"
"reducedUniqueWCNFs/MaxHS_60_red-3683434563126480531.wcnf*306885689"
"reducedUniqueWCNFs/maxpre2EvalMaxSAT_13_red-1787893125344086754.wcnf*1"
"reducedUniqueWCNFs/maxpre2EvalMaxSAT_13_red-452785039954645480.wcnf*2"
"reducedUniqueWCNFs/maxpre2EvalMaxSAT_13_red-589619030201996552.wcnf*1"
"reducedUniqueWCNFs/maxpre2EvalMaxSAT_13_red-858808193045986970.wcnf*1"
"reducedUniqueWCNFs/maxpre2EvalMaxSAT_39_red-1011490753210574616.wcnf*632536388"
"reducedUniqueWCNFs/maxpre2EvalMaxSAT_39_red-1964975267777150992.wcnf*1"
"reducedUniqueWCNFs/maxpre2EvalMaxSAT_43_red-2868335317395646542.wcnf*2"
"reducedUniqueWCNFs/maxpre2EvalMaxSAT_43_red-3006442318390157868.wcnf*2"
"reducedUniqueWCNFs/maxpre2EvalMaxSAT_43_red-4163859177505608977.wcnf*746423467"
"reducedUniqueWCNFs/maxpre2EvalMaxSAT_43_red-983158826512621529.wcnf*2148729523"
"reducedUniqueWCNFs/maxpre2EvalMaxSAT_56_red-1671103287372494433.wcnf*0"
"reducedUniqueWCNFs/maxpre2EvalMaxSAT_56_red-2503427328718924818.wcnf*0"
"reducedUniqueWCNFs/maxpre2EvalMaxSAT_56_red-2768918628154372673.wcnf*0"
"reducedUniqueWCNFs/maxpre2EvalMaxSAT_56_red-4599407379028638546.wcnf*0"
"reducedUniqueWCNFs/maxpre2EvalMaxSAT_9_red-340722805977130735.wcnf*0"
"reducedUniqueWCNFs/maxpre2EvalMaxSAT_9_red-831546485347623560.wcnf*2"
"reducedUniqueWCNFs/Pacose_12_red-1988129651103645923.wcnf*45"
"reducedUniqueWCNFs/Pacose_12_red-2269685325662887712.wcnf*48"
"reducedUniqueWCNFs/Pacose_12_red-3401324576103794552.wcnf*5"
"reducedUniqueWCNFs/Pacose_12_red-408069156423875169.wcnf*0"
"reducedUniqueWCNFs/Pacose_24_red-1286162711923734785.wcnf*17"
"reducedUniqueWCNFs/Pacose_24_red-1513046323290504422.wcnf*12"
"reducedUniqueWCNFs/Pacose_24_red-2697442186199561910.wcnf*72"
"reducedUniqueWCNFs/Pacose_24_red-3303886508067122356.wcnf*141"
"reducedUniqueWCNFs/Pacose_26_red-3151401845094493228.wcnf*0"
"reducedUniqueWCNFs/Pacose_26_red-3462212584067557464.wcnf*0"
"reducedUniqueWCNFs/Pacose_26_red-3988839199503314461.wcnf*0"
"reducedUniqueWCNFs/Pacose_30_red-2167694050811453563.wcnf*36"
"reducedUniqueWCNFs/Pacose_30_red-2192125470958684693.wcnf*115"
"reducedUniqueWCNFs/Pacose_30_red-2815432333344935313.wcnf*27"
"reducedUniqueWCNFs/Pacose_30_red-3941136238163041341.wcnf*25"
# 2^64-1, too big number -- "reducedUniqueWCNFs/Pacose_31_red-1341806195541177444.wcnf*UNSATISFIABLE" 
# 2^64-1, too big number -- "reducedUniqueWCNFs/Pacose_31_red-2886981514891097060.wcnf*8273860297810391990"
# 2^64-1, too big number -- "reducedUniqueWCNFs/Pacose_31_red-2959049275374198388.wcnf*6779779821486740930"
"reducedUniqueWCNFs/Pacose_35_red-2951893180766256529.wcnf*91398525"
"reducedUniqueWCNFs/Pacose_35_red-3233853261563692587.wcnf*1"
"reducedUniqueWCNFs/Pacose_35_red-930801946434130515.wcnf*1031300611"
# 2^64-1, too big number -- "reducedUniqueWCNFs/Pacose_39_red-1209140347810141628.wcnf*UNSATISFIABLE"
# 2^64-1, too big number -- "reducedUniqueWCNFs/Pacose_39_red-1332588873097487382.wcnf*UNSATISFIABLE"
# 2^64-1, too big number -- "reducedUniqueWCNFs/Pacose_39_red-1342887720528763864.wcnf*UNSATISFIABLE"
# 2^64-1, too big number -- "reducedUniqueWCNFs/Pacose_39_red-3740084684244600160.wcnf*UNSATISFIABLE"
# 2^64-1, too big number -- "reducedUniqueWCNFs/Pacose_42_red-2756940703361207309.wcnf*9698162971219944080"
# 2^64-1, too big number -- "reducedUniqueWCNFs/Pacose_42_red-3284435227402927186.wcnf*1698901850534264111"
# 2^64-1, too big number -- "reducedUniqueWCNFs/Pacose_42_red-3471727395647232998.wcnf*2028845280456358810"
#"reducedUniqueWCNFs/Pacose_42_red-4002787259383059368.wcnf*14139646852149208145" # sum of weights > 2^63
# 2^64-1, too big number -- "reducedUniqueWCNFs/Pacose_42_red-967626118846512481.wcnf*367634612884852860"
# 2^64-1, too big number -- "reducedUniqueWCNFs/Pacose_45_red-18861195512333290.wcnf*9257497295716810972"
# 2^64-1, too big number -- "reducedUniqueWCNFs/Pacose_45_red-2110119198975958323.wcnf*9445776242008088885"
# 2^64-1, too big number -- "reducedUniqueWCNFs/Pacose_45_red-4420657514169022274.wcnf*11195583190357724074"
# 2^64-1, too big number -- "reducedUniqueWCNFs/Pacose_45_red-692274117469184565.wcnf*10639677794134835561"
# 2^64-1, too big number -- "reducedUniqueWCNFs/Pacose_45_red-96855392154937795.wcnf*10546856073964118777"
# 2^64-1, too big number -- "reducedUniqueWCNFs/Pacose_48_red-1714475548514705799.wcnf*1558686677985956148"
# 2^64-1, too big number -- "reducedUniqueWCNFs/Pacose_50_red-1398795640098580938.wcnf*UNSATISFIABLE"
# 2^64-1, too big number -- "reducedUniqueWCNFs/Pacose_50_red-1537816876256169324.wcnf*UNSATISFIABLE"
# 2^64-1, too big number -- "reducedUniqueWCNFs/Pacose_50_red-1933202413033182321.wcnf*UNSATISFIABLE"
# 2^64-1, too big number -- "reducedUniqueWCNFs/Pacose_50_red-3191921006271149668.wcnf*UNSATISFIABLE"
# 2^64-1, too big number -- "reducedUniqueWCNFs/Pacose_50_red-3979338603693287500.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/Pacose_54_red-1358157514203883747.wcnf*26426196704"
"reducedUniqueWCNFs/Pacose_54_red-1649039155818303700.wcnf*26394559336"
"reducedUniqueWCNFs/Pacose_54_red-29072564938520591.wcnf*31370776965"
"reducedUniqueWCNFs/Pacose_54_red-482873502707281785.wcnf*44924798047"
"reducedUniqueWCNFs/Pacose_54_red-689259032191887290.wcnf*40623749314"
"reducedUniqueWCNFs/Pacose_56_red-3281457200392413701.wcnf*2223671546126634343"
#"reducedUniqueWCNFs/Pacose_56_red-4009874334574485555.wcnf*3810009383098323670" # sum of weights > 2^63
"reducedUniqueWCNFs/Pacose_56_red-4143592419272789868.wcnf*4473887333319298517"
"reducedUniqueWCNFs/Pacose_56_red-985471373479038083.wcnf*1815207463649088343"
"reducedUniqueWCNFs/Pacose_5_red-3303939259103370123.wcnf*1721"
"reducedUniqueWCNFs/Pacose_5_red-334040274351493152.wcnf*3167248"
"reducedUniqueWCNFs/Pacose_5_red-3802828968457918065.wcnf*54047"
"reducedUniqueWCNFs/UWrMaxSat_18_red-392707243128567773.wcnf*42"
"reducedUniqueWCNFs/UWrMaxSat_18_red-434222651361422376.wcnf*62"
#"reducedUniqueWCNFs/UWrMaxSat_48_red-3345226870336081251.wcnf*2666215797507687042"
#"reducedUniqueWCNFs/UWrMaxSat_48_red-4239671311353456234.wcnf*2358887300941932123"
#"reducedUniqueWCNFs/UWrMaxSat_48_red-4475012773486397200.wcnf*1238832899796780742"
#"reducedUniqueWCNFs/UWrMaxSat_48_red-527025851555882477.wcnf*1392571968806640307"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_12_red-2183858250570005557.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_12_red-2378267317391590079.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_12_red-2647874771286737686.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_12_red-708605192798371900.wcnf*1"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_14_red-1910526412935573211.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_14_red-2871396909313594.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_14_red-3062830906459114413.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_14_red-641133889199110692.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_18_red-2180853663016225873.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_18_red-2702238238001109781.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_18_red-2839957876382344943.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_18_red-3728880575233737753.wcnf*0"
#"reducedUniqueWCNFs/UWrMaxSat-SCIP_33_red-1069423359933276817.wcnf*9239063301820962052" # sum of weights > 2^63
#"reducedUniqueWCNFs/UWrMaxSat-SCIP_33_red-2320692503609130440.wcnf*9233114895876828182" # sum of weights > 2^63
#"reducedUniqueWCNFs/UWrMaxSat-SCIP_33_red-4489510176210977570.wcnf*9267979088012949242" # sum of weights > 2^63
#"reducedUniqueWCNFs/UWrMaxSat-SCIP_33_red-4604869691953281237.wcnf*9229160939207881892" # sum of weights > 2^63
#"reducedUniqueWCNFs/UWrMaxSat-SCIP_33_red-864786421590678020.wcnf*9269297536667740589" # sum of weights > 2^63
"reducedUniqueWCNFs/UWrMaxSat-SCIP_39_red-185185273772798505.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_39_red-244821460761367378.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_39_red-3091341505080918371.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_39_red-52396222493922010.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_42_red-1120907567350714328.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_42_red-1553028105115706290.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_42_red-1609334727346068066.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_42_red-2206054070726780272.wcnf*945747779"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_44_red-3281457200392413701.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_44_red-4143592419272789868.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_44_red-729035501448355466.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_44_red-985471373479038083.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_45_red-2603710192939071098.wcnf*1"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_45_red-4179909228510409015.wcnf*15127684839857653"
#"reducedUniqueWCNFs/UWrMaxSat-SCIP_45_red-4315342728986121571.wcnf*9298397313969497932" # sum of weights > 2^63
"reducedUniqueWCNFs/UWrMaxSat-SCIP_45_red-4342119793509242537.wcnf*1"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_48_red-1093108851438549088.wcnf*1650022201"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_48_red-212693676053199578.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_48_red-2965527567265231506.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_48_red-4414327767837536607.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_50_red-2497243989110826661.wcnf*1"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_50_red-2721944668192688995.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_50_red-3459617580742672707.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_50_red-3494417798159495887.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_50_red-719310437759088578.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_59_red-1270251658374345229.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_59_red-2174981755726662166.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_59_red-249136647606697574.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_59_red-4252545608530144281.wcnf*0"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_9_red-1168521650876440106.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_9_red-3135556796789662033.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_9_red-677120424724792323.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/UWrMaxSat-SCIP_9_red-827404868206741790.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/WMaxCDCL_31_red-2853693958521466174.wcnf*16906386530742218"
"reducedUniqueWCNFs/WMaxCDCL_31_red-2956309793569275881.wcnf*14081818676797022"
"reducedUniqueWCNFs/WMaxCDCL_31_red-3099223885584927630.wcnf*9467658688642377"
"reducedUniqueWCNFs/WMaxCDCL_39_red-1127604550348622587.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/WMaxCDCL_39_red-2180273152100062537.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/WMaxCDCL_39_red-3760685081229858521.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/WMaxCDCL_40_red-1230756892310092431.wcnf*3380361619521023"
"reducedUniqueWCNFs/WMaxCDCL_40_red-773616030500625105.wcnf*1480436609332498"
"reducedUniqueWCNFs/WMaxCDCL_42_red-1190973438965526534.wcnf*9266157091273657"
"reducedUniqueWCNFs/WMaxCDCL_42_red-2433556937788437875.wcnf*3"
"reducedUniqueWCNFs/WMaxCDCL_42_red-353414103522320368.wcnf*9491922604943463"
"reducedUniqueWCNFs/WMaxCDCL_44_red-3247157704190641489.wcnf*14113912232380779"
"reducedUniqueWCNFs/WMaxCDCL_44_red-3837120235180337223.wcnf*129800595045194515"
"reducedUniqueWCNFs/WMaxCDCL_45_red-3683220712223142792.wcnf*9514208105035831"
"reducedUniqueWCNFs/WMaxCDCL_45_red-4504257650829621829.wcnf*9319692811377733"
"reducedUniqueWCNFs/WMaxCDCL_48_red-2726655741059448275.wcnf*22201918151458737"
"reducedUniqueWCNFs/WMaxCDCL_60_red-3269417686496778418.wcnf*356393726267587704"
"reducedUniqueWCNFs/WMaxCDCL_60_red-3297696176384782222.wcnf*10204071671410622"
"reducedUniqueWCNFs/WMaxCDCL_60_red-3962628852343967740.wcnf*4221028885982231"
"reducedUniqueWCNFs/WMaxCDCL_9_red-3403352658369124999.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/WMaxCDCL_9_red-960852779331877826.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_14_red-1573595701574387285.wcnf*2"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_14_red-2607927524698267685.wcnf*4"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_14_red-2677058510767564796.wcnf*2"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_14_red-42326731340520793.wcnf*4"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_15_red-1195220942049486016.wcnf*3"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_15_red-2807986894681478028.wcnf*4"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_15_red-319672805271844931.wcnf*2"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_15_red-4073180598204406524.wcnf*7"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_15_red-4464408172294026275.wcnf*5"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_31_red-1650492680461762036.wcnf*9094379793822221"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_31_red-4582387355587878791.wcnf*8217076703818126"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_34_red-1580105472501737012.wcnf*68416055126301211"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_34_red-1743280836669036903.wcnf*13311291387918357"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_34_red-2383889806628104396.wcnf*12744948911465366"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_34_red-3910947893840279638.wcnf*8145900349697497"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_39_red-1130975699679110928.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_39_red-3806958441267743866.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_39_red-396029718659683991.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_39_red-821310142572620750.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_40_red-1136848343093114422.wcnf*2845007724092938"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_40_red-1682766907044057079.wcnf*1516626818063261321"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_40_red-4143592419272789868.wcnf*4665790745889012"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_40_red-4413781331871572532.wcnf*1329398780228330"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_42_red-1016238135184499311.wcnf*2"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_42_red-2287946345460757178.wcnf*9372883804288689"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_42_red-2665771636720740588.wcnf*9236471880757421"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_42_red-4093007172495955011.wcnf*2"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_44_red-1063503251099249683.wcnf*42791905684320438"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_44_red-1173477614390252516.wcnf*179949763010381823"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_44_red-247591745634806272.wcnf*2067324956135439479"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_44_red-2672274648256964035.wcnf*9125479794677607"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_45_red-2258465440550850344.wcnf*12389995001027008"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_45_red-2387742230118646907.wcnf*9268551008951731"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_45_red-935827160941047007.wcnf*9451751008398681"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_45_red-946033746336718185.wcnf*9149857280786209"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_48_red-1636324227788125146.wcnf*108803809607481811"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_48_red-3669774097419431089.wcnf*39936775820971482"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_48_red-454510086847019100.wcnf*187399513858102078"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_48_red-832511782895553194.wcnf*242000448646113205"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_56_red-4261954262327454570.wcnf*237803482746857698"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_56_red-645529805559806344.wcnf*227214827879778334"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_60_red-274178739103012090.wcnf*896473157272305"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_60_red-2775179874477784387.wcnf*8288849281776061"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_60_red-4089224804280863269.wcnf*13541078493145401"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_60_red-786798723718669031.wcnf*4157178231546598"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_9_red-2042077545131207184.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_9_red-2984188375622753476.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_9_red-706680171868523848.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/WMaxCDCL-BandAll_9_red-960852779331877826.wcnf*UNSATISFIABLE"
"reducedUniqueWCNFs/z3maxres_12_red-2690444265383616466.wcnf*4"
"reducedUniqueWCNFs/z3maxres_12_red-4442551375815625515.wcnf*4"
"reducedUniqueWCNFs/z3maxres_18_red-1315352413729120633.wcnf*1"
"reducedUniqueWCNFs/z3maxres_31_red-30358273458052886.wcnf*1882211714"
"reducedUniqueWCNFs/z3maxres_40_red-1682766907044057079.wcnf*2906095073836"
"reducedUniqueWCNFs/z3maxres_40_red-2678122237128123310.wcnf*693836500"
"reducedUniqueWCNFs/z3maxres_42_red-13898772362283218.wcnf*97526002455"
"reducedUniqueWCNFs/z3maxres_42_red-3551257043634425947.wcnf*1"
"reducedUniqueWCNFs/z3maxres_42_red-786798723718669031.wcnf*508967635"
"reducedUniqueWCNFs/z3maxres_45_red-1544036629320291585.wcnf*4453186682"
"reducedUniqueWCNFs/z3maxres_45_red-808243178418896134.wcnf*4520272313"
"reducedUniqueWCNFs/z3maxres_48_red-2173884515872664178.wcnf*63581687"
"reducedUniqueWCNFs/z3maxres_48_red-450710543792650697.wcnf*885191029"
"reducedUniqueWCNFs/z3maxres_48_red-514898981072592001.wcnf*2160415111"
"reducedUniqueWCNFs/z3maxres-bin_12_red-2690444265383616466.wcnf*4"
"reducedUniqueWCNFs/z3maxres-bin_12_red-3194397308415094363.wcnf*1"
"reducedUniqueWCNFs/z3maxres-bin_12_red-3441939117536983949.wcnf*4"
"reducedUniqueWCNFs/z3maxres-bin_12_red-419638518716462176.wcnf*4"
"reducedUniqueWCNFs/z3maxres-bin_12_red-4442551375815625515.wcnf*3"
"reducedUniqueWCNFs/z3maxres-bin_15_red-1240084053620135024.wcnf*4"
"reducedUniqueWCNFs/z3maxres-bin_15_red-1559107368525074383.wcnf*3"
"reducedUniqueWCNFs/z3maxres-bin_15_red-1975568268293811615.wcnf*4"
"reducedUniqueWCNFs/z3maxres-bin_15_red-2250018949800838538.wcnf*5"
"reducedUniqueWCNFs/z3maxres-bin_15_red-2301155001538167641.wcnf*3"
"reducedUniqueWCNFs/z3maxres-bin_18_red-2464473109497482716.wcnf*3"
"reducedUniqueWCNFs/z3maxres-bin_18_red-3260574134640532552.wcnf*4"
"reducedUniqueWCNFs/z3maxres-bin_18_red-3562335048339366804.wcnf*4"
"reducedUniqueWCNFs/z3maxres-bin_18_red-747533574540503553.wcnf*4"
"reducedUniqueWCNFs/z3maxres-bin_40_red-2078584792271772246.wcnf*165432189"
"reducedUniqueWCNFs/z3maxres-bin_40_red-2279939624383482426.wcnf*908930632"
"reducedUniqueWCNFs/z3maxres-bin_40_red-3883978496164186573.wcnf*975795372"
"reducedUniqueWCNFs/z3maxres-bin_42_red-1220543553353367910.wcnf*80059114"
"reducedUniqueWCNFs/z3maxres-bin_42_red-1230756892310092431.wcnf*445833420840687"
"reducedUniqueWCNFs/z3maxres-bin_42_red-4413781331871572532.wcnf*12270064718644"
"reducedUniqueWCNFs/z3maxres-bin_42_red-942966585981461076.wcnf*67512866452"
"reducedUniqueWCNFs/z3maxres-bin_45_red-1823565826937164294.wcnf*4307539575"
"reducedUniqueWCNFs/z3maxres-bin_45_red-2216804906134425387.wcnf*4456875748"
"reducedUniqueWCNFs/z3maxres-bin_45_red-2270454511839139899.wcnf*4558904152"
"reducedUniqueWCNFs/z3maxres-bin_45_red-2921088884541901271.wcnf*4465285311"
"reducedUniqueWCNFs/z3maxres-bin_48_red-1084762891037370712.wcnf*1246848185"
"reducedUniqueWCNFs/z3maxres-bin_48_red-1493345390056425047.wcnf*1"
"reducedUniqueWCNFs/z3maxres-bin_48_red-2398977464047622235.wcnf*1344913545"
"reducedUniqueWCNFs/z3pd-maxres_12_red-3194397308415094363.wcnf*1"
"reducedUniqueWCNFs/z3pd-maxres_12_red-3686736779330817673.wcnf*1"
"reducedUniqueWCNFs/z3pd-maxres_12_red-419638518716462176.wcnf*4"
"reducedUniqueWCNFs/z3pd-maxres_12_red-4563567611226058507.wcnf*1"
"reducedUniqueWCNFs/z3pd-maxres_15_red-1240084053620135024.wcnf*3"
"reducedUniqueWCNFs/z3pd-maxres_15_red-1559107368525074383.wcnf*3"
"reducedUniqueWCNFs/z3pd-maxres_15_red-1975568268293811615.wcnf*4"
"reducedUniqueWCNFs/z3pd-maxres_15_red-2250018949800838538.wcnf*5"
"reducedUniqueWCNFs/z3pd-maxres_15_red-2301155001538167641.wcnf*4"
"reducedUniqueWCNFs/z3pd-maxres_18_red-1315352413729120633.wcnf*1"
"reducedUniqueWCNFs/z3pd-maxres_18_red-2464473109497482716.wcnf*4"
"reducedUniqueWCNFs/z3pd-maxres_18_red-775858539240550471.wcnf*3"
"reducedUniqueWCNFs/z3pd-maxres_18_red-868561801701256536.wcnf*4"
"reducedUniqueWCNFs/z3pd-maxres_31_red-30358273458052886.wcnf*352167398"
"reducedUniqueWCNFs/z3pd-maxres_40_red-106264904244432591.wcnf*60819237"
"reducedUniqueWCNFs/z3pd-maxres_40_red-1136848343093114422.wcnf*1341057363"
"reducedUniqueWCNFs/z3pd-maxres_40_red-1411231300624613679.wcnf*761604759"
"reducedUniqueWCNFs/z3pd-maxres_40_red-3073507399860159194.wcnf*725285612"
"reducedUniqueWCNFs/z3pd-maxres_42_red-3269417686496778418.wcnf*5560109"
"reducedUniqueWCNFs/z3pd-maxres_42_red-4143592419272789868.wcnf*69593670145047"
"reducedUniqueWCNFs/z3pd-maxres_42_red-4413781331871572532.wcnf*12270064718644"
"reducedUniqueWCNFs/z3pd-maxres_42_red-729035501448355466.wcnf*175504879"
"reducedUniqueWCNFs/z3pd-maxres_45_red-1224009853503763696.wcnf*4574977835"
"reducedUniqueWCNFs/z3pd-maxres_45_red-1823565826937164294.wcnf*4307539575"
"reducedUniqueWCNFs/z3pd-maxres_45_red-3609484220625468693.wcnf*4352976280"
"reducedUniqueWCNFs/z3pd-maxres_45_red-4295153604889488190.wcnf*4338607289"
"reducedUniqueWCNFs/z3pd-maxres_48_red-1360629559721635608.wcnf*930972628"
"reducedUniqueWCNFs/z3pd-maxres_48_red-2173884515872664178.wcnf*194523753"
"reducedUniqueWCNFs/z3pd-maxres_48_red-2398977464047622235.wcnf*1344913545"
"reducedUniqueWCNFs/z3pd-maxres_48_red-3410017302197439143.wcnf*392410907"
"reducedUniqueWCNFs/z3rc2_12_red-2296016229024490956.wcnf*2"
"reducedUniqueWCNFs/z3rc2_12_red-2350789155647613363.wcnf*2"
"reducedUniqueWCNFs/z3rc2_12_red-3098761524201758363.wcnf*32"
"reducedUniqueWCNFs/z3rc2_12_red-4169755519552890842.wcnf*1"
"reducedUniqueWCNFs/z3rc2_15_red-1240084053620135024.wcnf*3"
"reducedUniqueWCNFs/z3rc2_15_red-1559107368525074383.wcnf*3"
"reducedUniqueWCNFs/z3rc2_18_red-175513428354106618.wcnf*2"
"reducedUniqueWCNFs/z3rc2_18_red-1975449769731875538.wcnf*1"
"reducedUniqueWCNFs/z3rc2_18_red-2005570631828580873.wcnf*5"
"reducedUniqueWCNFs/z3rc2_18_red-340722805977130735.wcnf*2"
"reducedUniqueWCNFs/z3rc2_40_red-2173826577747340621.wcnf*482764327"
"reducedUniqueWCNFs/z3rc2_40_red-2203871708166975328.wcnf*2868382509"
"reducedUniqueWCNFs/z3rc2_40_red-3565704524217356983.wcnf*513724829"
"reducedUniqueWCNFs/z3rc2_40_red-4504257650829621829.wcnf*8594142881685"
"reducedUniqueWCNFs/z3rc2_42_red-2030875441587988817.wcnf*1772420315"
"reducedUniqueWCNFs/z3rc2_42_red-3795776497734084734.wcnf*2398549951"
"reducedUniqueWCNFs/z3rc2_42_red-3928857511080506899.wcnf*886578922"
"reducedUniqueWCNFs/z3rc2_42_red-4413781331871572532.wcnf*12270064718644"
"reducedUniqueWCNFs/z3rc2_45_red-1823565826937164294.wcnf*4307539575"
"reducedUniqueWCNFs/z3rc2_45_red-329655784309346590.wcnf*4513951901"
"reducedUniqueWCNFs/z3rc2_45_red-3683220712223142792.wcnf*4536727956"
"reducedUniqueWCNFs/z3rc2_45_red-946033746336718185.wcnf*4305529855"
"reducedUniqueWCNFs/z3rc2_48_red-2939344604895055236.wcnf*465872227"
"reducedUniqueWCNFs/z3rc2_48_red-329610292110303123.wcnf*1688480088"
"reducedUniqueWCNFs/z3rc2_48_red-3561159528770161350.wcnf*299615600"
"reducedUniqueWCNFs/z3rc2_48_red-514898981072592001.wcnf*1"
"reducedUniqueWCNFs/z3wmax_23_red-2478727022419252999.wcnf*0"
"reducedUniqueWCNFs/z3wmax_23_red-340722805977130735.wcnf*0"
"reducedUniqueWCNFs/z3wmax_23_red-4520349575283451183.wcnf*0"
"reducedUniqueWCNFs/z3wmax_39_red-1957787610028959388.wcnf*0"
"reducedUniqueWCNFs/z3wmax_39_red-2381999570689558720.wcnf*0"
"reducedUniqueWCNFs/z3wmax_39_red-3142626626847041320.wcnf*0"
"reducedUniqueWCNFs/z3wmax_39_red-4500237089356291910.wcnf*4364620882"
"reducedUniqueWCNFs/z3wmax_40_red-1411231300624613679.wcnf*761604759"
"reducedUniqueWCNFs/z3wmax_40_red-1825580406392377719.wcnf*69909816816"
"reducedUniqueWCNFs/z3wmax_40_red-2078584792271772246.wcnf*165432189"
"reducedUniqueWCNFs/z3wmax_40_red-2735645343719678363.wcnf*1482755547"
"reducedUniqueWCNFs/z3wmax_42_red-1639582119976648487.wcnf*46022862304192"
"reducedUniqueWCNFs/z3wmax_42_red-2831519557877346324.wcnf*716289530052780"
"reducedUniqueWCNFs/z3wmax_42_red-287864067648591933.wcnf*117287631"
"reducedUniqueWCNFs/z3wmax_42_red-4268386316085429275.wcnf*874111437458710"
"reducedUniqueWCNFs/z3wmax_43_red-1757173891391360881.wcnf*1"
"reducedUniqueWCNFs/z3wmax_43_red-3711503600115445803.wcnf*0"
"reducedUniqueWCNFs/z3wmax_43_red-3987733112519799171.wcnf*0"
"reducedUniqueWCNFs/z3wmax_43_red-419516046229911735.wcnf*0"
"reducedUniqueWCNFs/z3wmax_45_red-2161187820670240308.wcnf*4572868525"
"reducedUniqueWCNFs/z3wmax_45_red-2439326252023750649.wcnf*4434859969"
"reducedUniqueWCNFs/z3wmax_45_red-3836323837818235591.wcnf*4448376598"
"reducedUniqueWCNFs/z3wmax_45_red-4161193175417292296.wcnf*4375681642"
"reducedUniqueWCNFs/z3wmax_48_red-140970586649036910.wcnf*1837985798"
"reducedUniqueWCNFs/z3wmax_48_red-2255722366088480985.wcnf*1369965139"
"reducedUniqueWCNFs/z3wmax_48_red-2259394261541393093.wcnf*529476472"
"reducedUniqueWCNFs/z3wmax_48_red-2480407866718150181.wcnf*239533808"
"reducedUniqueWCNFs/z3wmax_48_red-3604028936772614171.wcnf*1243373556"
"reducedUniqueWCNFs/z3wmax_53_red-1084762891037370712.wcnf*0"
"reducedUniqueWCNFs/z3wmax_53_red-2265738443285145225.wcnf*0"
"reducedUniqueWCNFs/z3wmax_53_red-3615035559607545761.wcnf*0"
"reducedUniqueWCNFs/z3wmax_53_red-4413781331871572532.wcnf*0"
)
fi

for j in "${wcnfArray[@]}"; do
    arraySize=${#wcnfArray[@]}
    formula="$(cut -d'*' -f1 <<<"$j")"
    formula="$SCRIPTPATH/$formula"
    if [ ! -f "$formula" ]; then
        echo "$formula does not exist."
        exit 1
    fi
    obj="$(cut -d'*' -f2 <<<"$j")"
    tested=$(( tested + 1 ))
    echo -ne "COUNTER: $tested / $arraySize --> $binary $formula            \r"
    # eval "$binary" "$formula"
    eval "$binary" "$formula" > output.log
    rv=$?

    if [ "$rv" != "0" ]; then
        if [ "$rv" != "10" ] && [ "$rv" != "20" ] && [ "$rv" != "30" ]; then
            echo ""
            echo "Error: Invalid return value: $rv"
            echo ""
            # echo "$binary $formula"
            errors=$(( errors + 1 ))
            continue
        fi
    fi

    value=$(awk '/^o/ {line=$0} /Error:/ || /UNSATISFIABLE/ || /UNKNOWN/ || /*Assertion*/ {print; line=""} END {print line}' output.log | cut -d' ' -f2)
    model=$(grep "^v " output.log)

    if [ -z "$value" ]; then
        emptyOvalue=$(( emptyOvalue + 1 ))
        echo ""
        echo "ERROR: empty o-value: $value"
        echo ""
        #echo "$binary $formula"
        continue
    fi

    if [ "$value" != "$obj" ]; then
        ovalueErrors=$(( ovalueErrors + 1 ))
        echo ""
        echo "ERROR: wrong o-value: $value vs expected $obj"
        echo ""
        #echo "$binary $formula"
        continue
    fi
    if [ -z "$model" ] && [ "$obj" != "UNSATISFIABLE" ]; then

        eval ./wcnfverifier -s ./kissat "$formula" > /dev/null
        rv=$?
        if [ "$rv" -ne "11" ]; then
          echo ""
          echo "Error: Hard Clauses are not satisfiable. RV of HardClauseVerifier: $rv"
          echo ./wcnfverifier -s ./kissat "$formula"
          echo ""
          #echo "$binary $formula"
          modelErrors=$(( modelErrors + 1 ))
          continue
        fi

        eval ./wcnfverifier -v output.log -o "$obj" "$formula" > /dev/null
        rv=$?
        if [ "$rv" != "0" ]; then
          echo "                                                                                                                                                                                         "
          echo "Error: wrong model. RV of ModelVerifier: $rv"
          echo "./wcnfverifier -v output.log -o $obj $formula"
          echo ""
          #echo "$binary $formula"
          modelErrors=$(( modelErrors + 1 ))
          continue
        fi
    fi
done

rm -f output.log

errorsum=$(( emptyOvalue + ovalueErrors + errors + modelErrors ))
echo "                                                                                                    "
echo "Invalid results.......: $errorsum / $tested"
echo "  no o value..........: $emptyOvalue"
echo "  wrong o value.......: $ovalueErrors"
echo "  invalid return value: $errors"
echo "  model errors........: $modelErrors"

if [ $errorsum -eq 0 ]; then
	exit 0
else
    exit 1
fi
