from veripb import InvalidProof
from veripb.rules import Rule, EmptyRule
from veripb.rules_register import register_rule
from veripb.parser import OPBParser, MaybeWordParser


@register_rule
class IsDeleted(EmptyRule):
    Ids = ["is_deleted"]

    @classmethod
    def parse(cls, line, context):
        cppWordIter = line.wordIter.getNative()
        ineq = context.ineqFactory.parse(cppWordIter, allowMultiple=False)
        found = context.propEngine.find(ineq, False)

        return cls(found)

    def __init__(self, found):
        self.found = found

    def compute(self, antecedents, context):
        if self.found:
            raise InvalidProof("Constraint should be deleted.")
        return []


@register_rule
class Fail(EmptyRule):
    Ids = ["fail"]

    @classmethod
    def parse(cls, line, context):
        return cls()

    def compute(self, antecedents, context):
        raise InvalidProof("Proof contains fail instruction.")
